#ifndef SMP_MPTCTL_GLUE_H
#define SMP_MPTCTL_GLUE_H

#ifndef __user
#define __user
#endif
/* Typedefs
 */
#ifndef uint
typedef unsigned int uint;
#endif

#ifndef ushort
typedef unsigned short ushort;
#endif

#ifndef uchar
typedef unsigned char uchar;
#endif

#ifndef u8
typedef unsigned char u8;
#endif

#ifndef u16
typedef unsigned short u16;
#endif

#ifndef u32
typedef unsigned int u32;
#endif

#ifndef u64
typedef unsigned long long u64;
#endif


#endif
